import hashlib
import itertools
import json
import os
from pathlib import Path
from typing import Any, Iterator, Self, Type

import msgspec
import packaging.requirements
import packaging.specifiers

from packse.template import load_template_config

type PackageName = str
type PackageVersion = (
    str  # TODO(zanieb): Consider replacing with `packaging.version.Version`
)


class Requirement(packaging.requirements.Requirement):
    """
    A wrapper for a standard Python requirement e.g. `foo>=1.0.0`
    """

    def __lt__(self, other: Any) -> True:
        """
        Sort by name for display purposes.
        """
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.name < other.name

    def with_unique_name(
        self,
        scenario: "Scenario",
        scenario_version: str,
        short_names: bool,
        no_hash: bool,
    ) -> type[Self]:
        """
        Return a copy of self with scenario metadata in the name
        """
        new = type(self)(str(self))
        new.name = self.name
        if not no_hash:
            new.name += "-" + scenario_version
        if not short_names:
            new.name = scenario.name + "-" + new.name
        return new


class PackageMetadata(msgspec.Struct, forbid_unknown_fields=True):
    """
    Metadata for a single version of a package.
    """

    requires_python: str | None = ">=3.8"
    requires: list[Requirement] = []
    extras: dict[str, list[Requirement]] = {}
    sdist: bool = True
    wheel: bool = True
    yanked: bool = False
    wheel_tags: list[str] = []
    description: str = ""

    def hash(self) -> str:
        """
        Return a hash of the contents
        """
        hasher = hashlib.new("md5", usedforsecurity=False)
        hasher.update((self.requires_python or "").encode())
        for requirement in self.requires:
            hasher.update(str(requirement).encode())
        for extra_name, requirements in self.extras.items():
            hasher.update(extra_name.encode())
            for requirement in requirements:
                hasher.update(str(requirement).encode())
        hasher.update(self.sdist.to_bytes())
        hasher.update(self.wheel.to_bytes())
        if self.yanked:
            hasher.update(self.yanked.to_bytes())
        if self.wheel:
            for wheel_tag in self.wheel_tags:
                hasher.update(wheel_tag.encode())
        hasher.update(self.description.encode())
        return hasher.hexdigest()

    def dict(self) -> dict:
        enc = msgspec.json.Encoder(enc_hook=enc_hook)
        return json.loads(enc.encode(self))


class RootPackageMetadata(msgspec.Struct, forbid_unknown_fields=True):
    requires_python: str | None = ">=3.8"
    requires: list[Requirement] = []

    def hash(self) -> str:
        """
        Return a hash of the contents
        """
        hasher = hashlib.new("md5", usedforsecurity=False)
        hasher.update((self.requires_python or "").encode())
        for requirement in self.requires:
            hasher.update(str(requirement).encode())
        return hasher.hexdigest()


class ResolverOptions(msgspec.Struct, forbid_unknown_fields=True):
    python: str | None = None
    """
    An optional Python version override.
    """

    prereleases: bool = False
    """
    If selection of prereleases should be enabled for all packages.
    """

    no_build: list[str] = []
    """
    Packages should be built from source distributions.

    Pre-built binaries are required for the given packages.
    """

    no_binary: list[str] = []
    """
    Pre-built binaries should not be allowed for the given packages.

    A source distribution is required to build the packages.
    """

    universal: bool = False
    """
    If resolution should operate in "universal" mode.

    In universal mode, a resolution can contain multiple versions of
    the same packages, but where different versions must be guarded by
    non-overlapping marker expressions.
    """

    def hash(self) -> str:
        """
        Return a hash of the contents
        """
        hasher = hashlib.new("md5", usedforsecurity=False)
        if self.python is not None:
            hasher.update(self.python.encode())
        if self.prereleases is not None:
            hasher.update(self.prereleases.to_bytes())
        for no_binary in self.no_binary:
            hasher.update(no_binary.encode())
        for no_build in self.no_build:
            hasher.update(no_build.encode())

        return hasher.hexdigest()


class EnvironmentMetadata(msgspec.Struct, forbid_unknown_fields=True):
    python: str = "3.8"
    """
    The active Python version.
    """

    additional_python: list[str] = []
    """
    Additional Python versions available on the system.

    By default, only the active Python version is available.
    """

    def hash(self) -> str:
        """
        Return a hash of the contents
        """
        hasher = hashlib.new("md5", usedforsecurity=False)
        hasher.update(self.python.encode())
        for additional_python in self.additional_python:
            hasher.update(additional_python.encode())
        return hasher.hexdigest()


class Package(msgspec.Struct, forbid_unknown_fields=True):
    versions: dict[PackageVersion, PackageMetadata]

    def hash(self) -> str:
        """
        Return a hash of the contents
        """
        hasher = hashlib.new("md5", usedforsecurity=False)
        for name, version in self.versions.items():
            hasher.update(name.encode())
            hasher.update(version.hash().encode())
        return hasher.hexdigest()


class Expected(msgspec.Struct, forbid_unknown_fields=True):
    """
    The expected outcome of a scenario.
    """

    satisfiable: bool
    """
    Are the packages in the scenario resolvable?
    """

    packages: dict[PackageName, PackageVersion] = {}
    """
    The expected versions to be installed for each package in the scenario.

    Only relevant if `satisfiable` is true.
    """

    explanation: str | None = None
    """
    An optional explanation for the expected result.
    """

    def hash(self) -> str:
        """
        Return a hash of the contents
        """
        hasher = hashlib.new("md5", usedforsecurity=False)
        hasher.update(str(self.satisfiable).encode())
        if self.explanation:
            hasher.update(self.explanation.encode())
        for name, package in self.packages.items():
            hasher.update(name.encode())
            hasher.update(package.encode())
        return hasher.hexdigest()


class Scenario(msgspec.Struct, forbid_unknown_fields=True):
    name: str
    """
    The name of the scenario.
    """

    packages: dict[PackageName, Package]
    """
    The packages available in the scenario.
    """

    root: RootPackageMetadata
    """
    The root package of the scenario.
    """

    expected: Expected
    """
    The expected result when resolving the scenario.
    """

    environment: EnvironmentMetadata = msgspec.field(
        default_factory=EnvironmentMetadata
    )
    """
    Metadata about the installation environment.
    """

    resolver_options: ResolverOptions = msgspec.field(default_factory=ResolverOptions)
    """
    Additional options for the package resolver
    """

    template: str = "package"
    """
    The template to use for scenario packages.
    """

    description: str | None = None
    """
    The description of the scenario.
    """

    def hash(self) -> str:
        """
        Return a hash of the scenario contents
        """
        hasher = hashlib.new("md5", usedforsecurity=False)
        hasher.update(self.name.encode())
        hasher.update(self.template.encode())
        hasher.update(self.root.hash().encode())
        hasher.update(self.environment.hash().encode())

        for name, package in self.packages.items():
            hasher.update(name.encode())
            hasher.update(package.hash().encode())

        # Note, the following are excluded because a re-publish is not needed on change
        # - resolver_options
        # - expected

        return hasher.hexdigest()

    def dict(self) -> dict:
        enc = msgspec.json.Encoder(enc_hook=enc_hook)
        return json.loads(enc.encode(self))


def _load(target: Path, type: Type):
    if target.suffix == ".json":
        return msgspec.json.decode(target.read_text(), type=type, dec_hook=dec_hook)
    elif target.suffix == ".toml":
        return msgspec.toml.decode(target.read_text(), type=type, dec_hook=dec_hook)
    elif target.suffix == ".yaml":
        return msgspec.yaml.decode(target.read_text(), type=type, dec_hook=dec_hook)
    else:
        raise ValueError(f"Unknown file type {target.suffix!r}")


def load_scenario(target: Path) -> Scenario:
    """
    Loads a scenario
    """
    return _load(target, type=Scenario)


def load_many_scenarios(target: Path) -> list[Scenario]:
    """
    Loads a file with many scenarios
    """
    return _load(target, type=list[Scenario])


def has_many_scenarios(target: Path) -> bool:
    # Guess if the file contains one or many scenario
    if target.suffix == ".json":
        with target.open() as buffer:
            return buffer.readline().lstrip().startswith("[")
    elif target.suffix == ".toml":
        return False  # Top-level arrays are not supported by TOML
    elif target.suffix == ".yaml":
        with target.open() as buffer:
            return buffer.readline().lstrip().startswith("-")


def load_scenarios(target: Path) -> list[Scenario]:
    if has_many_scenarios(target):
        return load_many_scenarios(target)
    else:
        return [load_scenario(target)]


def find_scenario_files(target: Path) -> Iterator[Path]:
    return itertools.chain(
        target.glob("**/*.json"), target.glob("**/*.yaml"), target.glob("**/*.toml")
    )


def scenario_hash(scenario: Scenario) -> str:
    """
    Generate a unique version for a scenario based on its contents.
    """
    template_version = load_template_config(scenario.template).version
    hasher = hashlib.new("md5", usedforsecurity=False)
    hasher.update(template_version.to_bytes())
    hasher.update(scenario.hash().encode())
    hasher.update(os.environ.get("PACKSE_VERSION_SEED", "").encode())
    return hasher.hexdigest()[:8]


def dec_hook(type: Type, obj: Any) -> Any:
    """
    Custom decoding hook for `Requirement` types.
    """
    # `type` here is the value of the custom type annotation being decoded.
    if type is Requirement:
        # Convert `obj` (which should be a `str`) to a `Requirement`
        return Requirement(obj)
    else:
        raise NotImplementedError(f"Objects of type {type} are not supported")


def enc_hook(obj: Any) -> Any:
    """
    Custom encoding hook for `Requirement` types
    """
    if isinstance(obj, Requirement):
        # convert the requirement to a str
        return str(obj)
    else:
        raise NotImplementedError(f"Objects of type {type(obj)} are not supported")



import logging
import shutil
import subprocess
import tempfile
import time
from pathlib import Path

import pkg_resources

from packse.error import DestinationAlreadyExists

logger = logging.getLogger(__name__)


def fetch(
    dest: Path | None = None,
    ref: str | None = None,
    repo_url: str = "https://github.com/astral-sh/packse",
    repo_subdir: str = "scenarios",
    force: bool = False,
):
    start_time = time.time()
    dest = dest or (Path.cwd() / "scenarios")

    if dest.exists() and not force:
        raise DestinationAlreadyExists(dest)

    if not ref:
        ref = pkg_resources.get_distribution("packse").version
        if ref == "0.0.0":
            ref = "HEAD"

    with tempfile.TemporaryDirectory() as tmpdir:
        # Perform a sparse checkout where we only grab the `scenarios` folder
        logger.info("Cloning repository %s", repo_url)
        subprocess.check_call(
            [
                "git",
                "clone",
                "-n",
                "--filter=tree:0",
                repo_url,
                "repo",
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            cwd=tmpdir,
        )

        logger.info("Checking out directory '%s' at ref %s", repo_subdir, ref)
        subprocess.check_call(
            ["git", "sparse-checkout", "set", "--no-cone", repo_subdir],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            cwd=Path(tmpdir) / "repo",
        )
        subprocess.check_call(
            ["git", "fetch", "origin", ref],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            cwd=Path(tmpdir) / "repo",
        )
        subprocess.check_call(
            ["git", "checkout", ref],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            cwd=Path(tmpdir) / "repo",
        )

        src = Path(tmpdir) / "repo" / repo_subdir
        file_count = 0
        for file in sorted(src.iterdir()):
            file_count += 1
            logger.info("Found %s", file.name)

        logger.info("Copying files into '%s'", dest)
        shutil.copytree(src, dest, dirs_exist_ok=True)

    logger.info(
        "Fetched %s files in %.2fs",
        file_count,
        time.time() - start_time,
    )



import importlib
import logging
import threading
from pathlib import Path

from packse.build import build
from packse.error import RequiresExtra, ServeThreadError
from packse.index import add_vendored_build_deps, start_index_server

try:
    import watchfiles
except ImportError:
    watchfiles = None
    pass


logger = logging.getLogger(__name__)


def serve(
    targets: list[Path],
    build_dir: Path,
    dist_dir: Path,
    host: str = "localhost",
    port: int = 3141,
    short_names: bool = False,
    no_hash: bool = False,
    offline: bool = False,
):
    try:
        importlib.import_module("pypiserver")
    except ImportError:
        raise RequiresExtra("index commands", "index")
    if watchfiles is None:
        raise RequiresExtra("serve command", "serve")

    dist_dir = Path(dist_dir)
    build_dir = Path(build_dir)

    build_ready = threading.Event()
    rebuild = threading.Thread(
        target=build_scenarios,
        args=(targets, short_names, no_hash, dist_dir, build_dir, build_ready),
        name="serve-build-scenarios",
        daemon=True,
    )
    serve = threading.Thread(
        target=serve_packages,
        args=(host, port, dist_dir, offline),
        name="serve-package-index",
    )
    rebuild.start()

    # Wait up to 60 seconds for the build to complete
    for _ in range(60):
        if build_ready.wait(1):
            break
        if not rebuild.is_alive():
            raise ServeThreadError("Build thread failed.")

    try:
        serve.start()
    finally:
        serve.join()


def build_scenarios(
    targets: list[Path],
    short_names: bool,
    no_hash: bool,
    dist_dir: Path,
    build_dir: Path,
    ready: threading.Event,
) -> None:
    if not targets:
        print("Warning: No targets provided, automatic rebuilds not enabled.")
    else:
        print("Performing initial build...")
        build(
            targets,
            rm_destination=True,
            skip_root=False,
            short_names=short_names,
            no_hash=no_hash,
            dist_dir=dist_dir,
            build_dir=build_dir,
        )

    ready.set()

    for changes in watchfiles.watch(*targets):
        targets = [path for kind, path in changes if kind != watchfiles.Change.deleted]
        targets = [Path(target) for target in targets if Path(target).is_file()]
        print("Detected changes! Rebuilding...")
        build(
            targets,
            rm_destination=True,
            skip_root=False,
            short_names=short_names,
            no_hash=no_hash,
            dist_dir=dist_dir,
            build_dir=build_dir,
        )


def serve_packages(host: str, port: int, dist_dir: Path, offline: bool):
    index_url = f"http://{host}:{port}"
    index_note = "offline index" if offline else "local index with PyPI fallback"

    logger.info("Starting %s at %s", index_note, index_url)
    with start_index_server(
        dist_dir=dist_dir, host=host, port=port, offline=offline, server_log_path=None
    ) as process:
        debug = logger.getEffectiveLevel() <= logging.DEBUG
        if not debug:
            logger.info("Hiding server output, use `-v` to stream server logs")

        add_vendored_build_deps(dist_dir, offline)

        logger.info("Ready! [Stop with Ctrl-C]")
        line = ""
        while True:
            line = process.stdout.readline().decode()
            if debug:
                print(line, end="")



"""
Build packages for the given scenarios.
"""

import logging
import shutil
import subprocess
import textwrap
import time
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait as wait_for_futures
from pathlib import Path
from typing import Generator

import packaging.version

from packse.error import (
    BuildError,
    DestinationAlreadyExists,
    FileNotFound,
    InvalidPackageVersion,
    InvalidScenario,
)
from packse.scenario import (
    Package,
    PackageMetadata,
    Scenario,
    find_scenario_files,
    load_scenarios,
    scenario_hash,
)
from packse.template import TemplateConfig, create_from_template, load_template_config

logger = logging.getLogger(__name__)


def build(
    targets: list[Path],
    rm_destination: bool,
    short_names: bool,
    no_hash: bool,
    skip_root: bool,
    dist_dir: Path | None = None,
    build_dir: Path | None = None,
):
    start_time = time.time()

    # Validate and collect all targets first
    scenarios = []
    dist_dir = dist_dir or (Path.cwd() / "dist")
    build_dir = build_dir or (Path.cwd() / "build")

    for target in targets:
        if not target.exists():
            raise FileNotFound(target)

        if target.is_dir():
            for target in find_scenario_files(target):
                try:
                    logger.debug("Loading %s", target)
                    scenarios.extend(load_scenarios(target))
                except Exception as exc:
                    invalid = InvalidScenario(target, reason=str(exc))
                    print(f"Skipping file: {invalid}")
        else:
            try:
                logger.debug("Loading %s", target)
                scenarios.extend(load_scenarios(target))
            except Exception as exc:
                raise InvalidScenario(target, reason=str(exc)) from exc

    # Then build each one
    with ThreadPoolExecutor(thread_name_prefix="packse-scenario-") as executor:
        futures = [
            executor.submit(
                build_scenario,
                scenario,
                rm_destination,
                short_names,
                no_hash,
                skip_root,
                build_dir,
                dist_dir,
            )
            for scenario in scenarios
        ]

        wait_for_futures(futures)

    results = [future.result() for future in futures]
    for result in sorted(results):
        print(result)

    logger.info(
        "Built %s scenarios in %.2fs",
        len(results),
        time.time() - start_time,
    )


def build_scenario(
    scenario: Scenario,
    rm_destination: bool,
    short_names: bool,
    no_hash: bool,
    skip_root: bool,
    build_dir: Path,
    dist_dir: Path,
) -> str:
    """
    Build the scenario defined at the given path.

    Returns the scenario's entrypoint package name.
    """

    hash = scenario_hash(scenario)
    name = scenario.name
    if not no_hash:
        name = f"{name}-{hash}"

    build_destination = build_dir / name
    dist_destination = dist_dir / name
    start_time = time.time()

    logger.info(
        "Building '%s' in directory '%s'",
        name,
        build_destination,
    )

    if build_destination.exists():
        if rm_destination:
            shutil.rmtree(build_destination)
        else:
            raise DestinationAlreadyExists(build_destination)

    build_destination.mkdir(parents=True)

    if dist_destination.exists():
        if rm_destination:
            shutil.rmtree(dist_destination)
        else:
            raise DestinationAlreadyExists(dist_destination)

    dist_destination.mkdir(parents=True)

    with ThreadPoolExecutor(thread_name_prefix="packse-build-") as executor:
        futures = [
            executor.submit(
                build_scenario_package,
                scenario=scenario,
                scenario_version=hash,
                name=name,
                package=package,
                build_destination=build_destination,
                dist_destination=dist_destination,
                short_names=short_names,
                no_hash=no_hash,
            )
            for name, package in scenario.packages.items()
        ]

        if not skip_root:
            futures.append(
                executor.submit(
                    build_scenario_package,
                    scenario=scenario,
                    scenario_version=hash,
                    name=scenario.name,
                    package=make_root_package(scenario),
                    build_destination=build_destination,
                    dist_destination=dist_destination,
                    short_names=short_names,
                    no_hash=no_hash,
                )
            )

        wait_for_futures(futures)
        for future in futures:
            # TODO(zanieb): Display _all_ of the errors to the user
            future.result()

    logger.info(
        "Built scenario '%s' in %.2fs",
        name,
        time.time() - start_time,
    )
    return name


def make_root_package(scenario: Scenario) -> Package:
    """
    Generate a full package from the root package of the scenario.
    """
    return Package(
        versions={
            "0.0.0": PackageMetadata(
                requires=scenario.root.requires,
                requires_python=scenario.root.requires_python,
                # Do not build wheels for the root package
                wheel=False,
                # The scenario's description is used for the entrypoint package
                description=scenario.description,
            )
        }
    )


def build_scenario_package(
    scenario: Scenario,
    scenario_version: str,
    name: str,
    package: Package,
    build_destination: Path,
    dist_destination: Path,
    short_names: bool,
    no_hash: bool,
):
    package_name = name
    if not no_hash:
        package_name = f"{package_name}-{scenario_version}"
    if not short_names and name != scenario.name:
        package_name = f"{scenario.name}-{package_name}"

    # Generate a Python module name
    module_name = package_name.replace("-", "_")

    template_config = load_template_config(scenario.template)

    for version, package_version in package.versions.items():
        start_time = time.time()

        logger.debug("Generating project for '%s-%s'", package_name, version)
        package_destination = create_from_template(
            build_destination,
            template_name=scenario.template,
            variables={
                "scenario-name": scenario.name,
                "package-name": package_name,
                "module-name": module_name,
                "version": version,
                "dependencies": [
                    requirement.with_unique_name(
                        scenario, scenario_version, short_names, no_hash
                    )
                    for requirement in package_version.requires
                ],
                "optional-dependencies": [
                    {
                        "name": extra,
                        "dependencies": [
                            requirement.with_unique_name(
                                scenario, scenario_version, short_names, no_hash
                            )
                            for requirement in depends
                        ],
                    }
                    for extra, depends in package_version.extras.items()
                ],
                "requires-python": package_version.requires_python,
                "description": package_version.description,
            },
        )

        logger.info(
            "Generated project for '%s-%s' in %.2fms",
            package_name,
            version,
            (time.time() - start_time) * 1000.0,
        )

        for dist in build_package_distributions(
            template_config, package_version, package_destination
        ):
            shared_path = dist_destination / dist.name
            logger.debug("Linked distribution to %s", shared_path)
            shared_path.hardlink_to(dist)

        if package_version.yanked:
            with dist_destination.with_suffix(".yanked").open("a+") as yanked:
                yanked.write(f"{package_name}-{version}")
                yanked.write("\n")
            logger.info("Marked %s-%s as yanked", package_name, version)


def build_package_distributions(
    template_config: TemplateConfig, package_version: PackageMetadata, target: Path
) -> Generator[Path, None, None]:
    """
    Build package distributions, yield each built distribution path, then delete the distribution folder.
    """
    command = template_config.build_base

    if package_version.sdist:
        command.extend(template_config.build_sdist)
    if package_version.wheel:
        command.extend(template_config.build_wheel)

    start_time = time.time()

    try:
        output = subprocess.check_output(command, cwd=target, stderr=subprocess.STDOUT)

    except subprocess.CalledProcessError as exc:
        raise BuildError(
            f"Building {target.name} with hatch failed",
            exc.output.decode(),
        )
    else:
        logs = (
            (":\n\n" + textwrap.indent(output.decode(), " " * 4))
            if logger.getEffectiveLevel() <= logging.DEBUG
            else ""
        )
        logger.info(
            "Built package '%s' in %.2fs%s",
            target.name,
            time.time() - start_time,
            logs,
        )

    # Create wheels with other tags if requested
    if package_version.wheel and package_version.wheel_tags:
        default_tag_wheel = None
        for dist in (target / "dist").iterdir():
            if dist.name.endswith("py3-none-any.whl"):
                default_tag_wheel = dist
                break
        if not default_tag_wheel:
            raise BuildError("No wheel found with tag `py3-none-any`.")

        # Since we always build a universal wheel, we just create copies of it for other platforms
        # which means we're lying about the compatibility of the wheel but it will always be installable
        # on the given platform.
        for tag in package_version.wheel_tags:
            if tag == "py3-none-any":
                continue

            new_tag_wheel = (
                target / "dist" / default_tag_wheel.name.replace("py3-none-any", tag)
            )
            new_tag_wheel.hardlink_to(default_tag_wheel)
            logger.debug("Created wheel %s", new_tag_wheel.name)

        # Delete the default wheel if not requested
        if "py3-none-any" not in package_version.wheel_tags:
            default_tag_wheel.unlink()

    yield from sorted((target / "dist").iterdir())
    shutil.rmtree(target / "dist")


def build_package(
    name: str,
    version: str,
    no_wheel: bool,
    no_sdist: bool,
    requires_python: str | None,
    wheel_tags: list[str],
    rm_destination: bool,
):
    """
    Build a package without a scenario
    """

    try:
        packaging.version.Version(version)
    except Exception:
        raise InvalidPackageVersion(version) from None

    work_dir = Path.cwd()
    build_destination = work_dir / "build" / name
    dist_destination = work_dir / "dist" / name
    start_time = time.time()

    if build_destination.exists():
        if rm_destination:
            shutil.rmtree(build_destination)
        else:
            raise DestinationAlreadyExists(build_destination)

    build_destination.mkdir(parents=True)

    if dist_destination.exists():
        if rm_destination:
            shutil.rmtree(dist_destination)
        else:
            raise DestinationAlreadyExists(dist_destination)

    dist_destination.mkdir(parents=True)
    package_name = name
    extra = {}
    if requires_python is not None:
        extra["requires_python"] = requires_python
    package_version = PackageMetadata(
        sdist=not no_sdist, wheel=not no_wheel, wheel_tags=wheel_tags, **extra
    )
    module_name = package_name.replace("-", "_")

    template_config = load_template_config("package")

    logger.debug("Generating project for '%s-%s'", package_name, version)

    package_destination = create_from_template(
        build_destination,
        template_name="package",
        variables={
            "package-name": package_name,
            "module-name": module_name,
            "version": version,
            "dependencies": [],
            "optional-dependencies": [],
            "requires-python": package_version.requires_python,
            "description": package_version.description,
        },
    )

    logger.info(
        "Generated project for '%s-%s'in %.2fms",
        package_name,
        version,
        (time.time() - start_time) * 1000.0,
    )

    for dist in build_package_distributions(
        template_config, package_version, package_destination
    ):
        shared_path = dist_destination / dist.name
        logger.info("Linked distribution to %s", shared_path.relative_to(work_dir))
        shared_path.hardlink_to(dist)

    logger.info(
        "Done in %.2fms",
        (time.time() - start_time) * 1000.0,
    )



import pytest

from .common import snapshot_command


def test_build_pkg_no_name(snapshot):
    assert snapshot_command(["build-pkg"]) == snapshot


def test_build_pkg_no_version(snapshot):
    assert snapshot_command(["build-pkg", "foo"]) == snapshot


def test_build_pkg_invalid_version(snapshot, tmpcwd):
    target = tmpcwd / "test.json"
    target.touch()
    assert snapshot_command(["build-pkg", "foo", "bar"]) == snapshot


@pytest.mark.usefixtures("tmpcwd")
def test_build_pkg(snapshot):
    assert (
        snapshot_command(
            ["build-pkg", "foo", "1.0.0"],
            snapshot_filesystem=True,
        )
        == snapshot
    )


@pytest.mark.usefixtures("tmpcwd")
def test_build_pkg_no_wheel(snapshot):
    assert (
        snapshot_command(
            ["build-pkg", "foo", "1.0.0", "--no-wheel"],
            snapshot_filesystem=True,
        )
        == snapshot
    )


@pytest.mark.usefixtures("tmpcwd")
def test_build_pkg_no_sdist(snapshot):
    assert (
        snapshot_command(
            ["build-pkg", "foo", "1.0.0", "--no-sdist"],
            snapshot_filesystem=True,
        )
        == snapshot
    )


@pytest.mark.usefixtures("tmpcwd")
def test_build_pkg_wheel_tags(snapshot):
    assert (
        snapshot_command(
            ["build-pkg", "foo", "1.0.0", "-t", "tag1", "--wheel-tag", "tag2"],
            snapshot_filesystem=True,
        )
        == snapshot
    )



import pytest

from .common import snapshot_command


def test_root_shows_help(snapshot):
    assert snapshot_command([]) == snapshot


@pytest.mark.parametrize("flag", ("-v", "--verbose"))
def test_root_verbose(snapshot, flag):
    assert snapshot_command([flag]) == snapshot



"""
View package tree for the given scenarios.
"""

import logging
from pathlib import Path

from packse.error import FileNotFound, InvalidScenario
from packse.scenario import (
    Package,
    Requirement,
    Scenario,
    load_scenarios,
    scenario_hash,
)

logger = logging.getLogger(__name__)


def view(targets: list[Path], name: str | None = None, short_names: bool = False):
    scenarios = []

    # Validate and collect all targets first
    for target in targets:
        if not target.exists():
            raise FileNotFound(target)

        try:
            logger.debug("Loading %s", target)
            scenarios.extend(load_scenarios(target))
        except Exception as exc:
            raise InvalidScenario(target, reason=str(exc)) from exc

    # Then view each one
    for scenario in scenarios:
        if (
            name is not None
            # Allow user to provide the name with or without the version / name
            and scenario.name != name
            and f"{scenario.name}-{scenario_hash(scenario)}" != name
            and scenario_hash(scenario) != name
        ):
            logging.debug("Skipping %s", scenario.name)
            continue

        # When viewing a single scenario, show the name and description
        if name is not None:
            print(scenario.name)
            print()
            print(scenario.description)
            print()
        else:
            logging.debug("Viewing %s", scenario.name)

        view_scenario(scenario, short_names)


def view_scenario(scenario: Scenario, short_names: bool):
    name = scenario_hash(scenario)
    if not short_names:
        name = f"{scenario.name}-{name}"

    print(name)
    print(dependency_tree(scenario))


def dependency_tree(scenario: Scenario) -> str:
    """
    Generate a dependency tree for a scenario
    """

    space = "    "
    branch = "│   "
    tee = "├── "
    last = "└── "
    buffer = ""

    packages = scenario.packages.copy()
    packages["root"] = Package(versions={"0.0.0": scenario.root})

    def render_versions(
        package: str,
        prefix: str = "",
        for_requirement: Requirement | None = None,
    ):
        versions = packages[package].versions
        if for_requirement:
            versions = {
                version: metadata
                for version, metadata in versions.items()
                if for_requirement.specifier.contains(version)
                if not metadata.yanked
            }

            if not versions:
                yield prefix + last + "unsatisfied: no matching version"
                return

        show_versions = []
        for version, metadata in versions.items():
            show_versions.append(
                (
                    version,
                    metadata.requires
                    + [Requirement(f"python{metadata.requires_python}")],
                    metadata,
                )
            )
            extras = metadata.extras

            for extra, extra_depends in extras.items():
                show_versions.append((f"{version}[{extra}]", extra_depends, metadata))

        pointers = [tee] * (len(show_versions) - 1) + [last]
        for pointer, (version, requirements, metadata) in zip(pointers, show_versions):
            message = "satisfied by " if for_requirement and package else ""
            yanked = " (yanked)" if metadata.yanked else ""
            yield prefix + pointer + message + f"{package}-{version}" + yanked

            if not for_requirement:
                extension = branch if pointer == tee else space
                yield from render_requirements(
                    requirements,
                    prefix=prefix + extension,
                )

    def render_requirements(requirements: list[Requirement], prefix: str = ""):
        for index, requirement in enumerate(requirements):
            # Skip display of Python requirements that are satisfied and not complex
            # We do this before the following iteration so the pointers are correct
            if (
                requirement.name == "python"
                and requirement.specifier.contains(scenario.environment.python)
                and len(requirement.specifier) == 1
                and next(iter(requirement.specifier)).version
                == scenario.environment.python
            ):
                requirements = requirements.copy()
                requirements.pop(index)

        pointers = [tee] * (len(requirements) - 1) + [last]
        for pointer, requirement in zip(pointers, sorted(requirements)):
            # Display `requires-python`
            if requirement.name == "python":
                suffix = (
                    " (incompatible with environment)"
                    if not requirement.specifier.contains(scenario.environment.python)
                    else ""
                )
                yield prefix + pointer + "requires " + str(requirement) + suffix
                continue

            yield prefix + pointer + "requires " + str(requirement)

            if requirement.name in packages:
                extension = branch if pointer == tee else space
                yield from render_versions(
                    requirement.name,
                    prefix=prefix + extension,
                    for_requirement=requirement,
                )
            else:
                yield prefix + space + last + "unsatisfied: no versions for package"

    # Print the environment first
    pointer = tee
    buffer += pointer + "environment\n"
    prefix = branch
    if scenario.environment.additional_python:
        python_versions = scenario.environment.additional_python + [
            scenario.environment.python
        ]
        pointers = [tee] * (len(python_versions) - 1) + [last]
        for pointer, version in zip(pointers, sorted(python_versions)):
            active = " (active)" if version == scenario.environment.python else ""
            buffer += prefix + pointer + f"python{version}" + active + "\n"
    else:
        buffer += prefix + last + f"python{scenario.environment.python}\n"

    # Print the root package first
    pointer = tee if scenario.packages else last
    buffer += pointer + "root\n"
    prefix = branch if pointer == tee else space

    # Then render versions for the root package
    for line in render_requirements(
        scenario.root.requires,
        prefix=prefix,
    ):
        buffer += line + "\n"

    # Then render all the other provided packages
    pointers = [tee] * (len(scenario.packages) - 1) + [last]
    for pointer, package in zip(pointers, sorted(scenario.packages)):
        # Print each package section
        buffer += pointer + package + "\n"

        prefix = branch if pointer == tee else space
        for line in render_versions(package, prefix=prefix):
            buffer += line + "\n"

    return buffer



"""
List all scenarios.
"""

import logging
from pathlib import Path
from typing import cast

from packse.error import FileNotFound, InvalidScenario
from packse.scenario import (
    Scenario,
    load_scenarios,
    scenario_hash,
)

logger = logging.getLogger(__name__)


def list(
    targets: list[Path],
    no_hash: bool = False,
    skip_invalid: bool = False,
    no_sources: bool = False,
    short_names: bool = False,
):
    scenarios_by_path: dict[Path, list[Scenario]] = {}

    # Validate and collect all targets first
    for target in sorted(targets):
        if not target.exists():
            raise FileNotFound(target)

        try:
            logger.debug("Loading %s", target)
            scenarios_by_path[target] = load_scenarios(target)
        except Exception as exc:
            if not skip_invalid:
                raise InvalidScenario(target, reason=str(exc)) from exc

    # Then list each one
    for source, scenarios in scenarios_by_path.items():
        prefix = "" if no_sources else " " * 4
        if not no_sources:
            print(to_display_path(source))

        for scenario in scenarios:
            scenario = cast(Scenario, scenario)

            if no_hash:
                name = scenario.name
            else:
                name = scenario_hash(scenario)
                if not short_names:
                    name = f"{scenario.name}-{name}"

            print(prefix + name)


def to_display_path(path: Path | str, relative_to: Path | str | None = None) -> str:
    """
    Convert a path to a displayable path. The absolute path or relative path to the
    current (or given) directory will be returned, whichever is shorter.
    """
    path, relative_to = (
        Path(path).resolve(),
        Path(relative_to or ".").resolve(),
    )

    absolute_path = str(path)

    try:
        relative_path = str(path.relative_to(relative_to))
    except ValueError:
        # Not a child path, use the other one
        return absolute_path

    return relative_path if len(relative_path) < len(absolute_path) else absolute_path



import os
import re
import shutil
import stat
import subprocess
import tempfile
from pathlib import Path
from typing import Generator

import pytest
from packse import __development_base_path__

from .common import snapshot_command


@pytest.fixture(scope="module")
def scenario_dist() -> Generator[Path, None, None]:
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"

    with tempfile.TemporaryDirectory() as tmpdir:
        subprocess.check_call(
            ["packse", "build", str(target)],
            cwd=tmpdir,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        dists = list((Path(tmpdir) / "dist").iterdir())
        assert len(dists) == 1
        dist = dists[0]

        yield dist


@pytest.fixture
def credentials(tmpenviron) -> None:
    os.environ["PACKSE_PUBLISH_USERNAME"] = "username"
    os.environ["PACKSE_PUBLISH_PASSWORD"] = "password"
    yield


class MockBinary:
    def __init__(self, path: Path) -> None:
        self.path = path
        self.callback = None
        self._update_bin("")

    def _prepare_text(self, text: str = None):
        if not text:
            return ""
        # Escape single quotes
        return text.replace("'", "\\'")

    def _update_bin(self, content: str):
        self.path.write_text("#!/usr/bin/env sh\n\n" + content + "\n")
        self.path.chmod(self.path.stat().st_mode | stat.S_IEXEC)

    def set_success(self, text: str | None = None):
        text = self._prepare_text(text)
        self._update_bin(f"echo '{text}'")

    def set_error(self, text: str | None = None):
        text = self._prepare_text(text)
        self._update_bin(f"echo '{text}'; exit 1")


@pytest.fixture
def mock_twine(monkeypatch: pytest.MonkeyPatch) -> Generator[MockBinary, None, None]:
    # Create a temp directory to register as a bin
    with tempfile.TemporaryDirectory() as tmpdir:
        mock = MockBinary(Path(tmpdir) / "twine")
        mock.set_success("<mock twine logs>")

        # Add to the path
        monkeypatch.setenv("PATH", tmpdir, prepend=":")
        assert shutil.which("twine").startswith(tmpdir)

        yield mock


@pytest.mark.usefixtures("credentials")
def test_publish_example_dry_run(snapshot, scenario_dist: Path):
    assert (
        snapshot_command(
            ["publish", "--dry-run", scenario_dist],
            extra_filters=[(re.escape(str(scenario_dist.resolve())), "[DISTDIR]")],
        )
        == snapshot
    )


@pytest.mark.usefixtures("credentials")
def test_publish_example_twine_succeeds(
    snapshot, scenario_dist: Path, mock_twine: MockBinary
):
    mock_twine.set_success("<twine happy message>")

    assert (
        snapshot_command(
            ["publish", scenario_dist, "-v", "--workers", "1"],
            extra_filters=[(re.escape(str(scenario_dist.resolve())), "[DISTDIR]")],
        )
        == snapshot
    )


@pytest.mark.usefixtures("credentials")
def test_publish_example_twine_succeeds_parallel(
    snapshot, scenario_dist: Path, mock_twine: MockBinary
):
    mock_twine.set_success("<twine happy message>")

    assert (
        snapshot_command(
            ["publish", scenario_dist, "-v"],
            extra_filters=[(re.escape(str(scenario_dist.resolve())), "[DISTDIR]")],
            # Cannot record stderr when running in parallel
            snapshot_stderr=False,
        )
        == snapshot
    )


@pytest.mark.usefixtures("credentials")
def test_publish_example_twine_fails_with_unknown_error(
    snapshot, scenario_dist: Path, mock_twine: MockBinary
):
    mock_twine.set_error("<twine error message>")

    assert (
        snapshot_command(
            ["publish", scenario_dist, "-v", "--workers", "1"],
            extra_filters=[(re.escape(str(scenario_dist.resolve())), "[DISTDIR]")],
        )
        == snapshot
    )


@pytest.mark.usefixtures("credentials")
def test_publish_example_twine_fails_with_rate_limit(
    snapshot, scenario_dist: Path, mock_twine: MockBinary
):
    mock_twine.set_error(
        """
Uploading distributions to https://test.pypi.org/legacy/
Uploading
requires_transitive_incompatible_with_root_version_5c1b7dc1_c-1.0.0-py3-none-any
.whl
25l
    0% ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━ 0.0/4.1 kB • --:-- • ?
100% ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━ 4.1/4.1 kB • 00:00 • ?
100% ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━ 4.1/4.1 kB • 00:00 • ?
100% ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━ 4.1/4.1 kB • 00:00 • ?
100% ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━ 4.1/4.1 kB • 00:00 • ?
25hWARNING  Error during upload. Retry with the --verbose option for more details.
ERROR    HTTPError: 429 Too Many Requests from https://test.pypi.org/legacy/
            Too many new projects created
        """
    )

    assert (
        snapshot_command(
            ["publish", scenario_dist, "--workers", "1"],
            extra_filters=[(re.escape(str(scenario_dist.resolve())), "[DISTDIR]")],
        )
        == snapshot
    )


@pytest.mark.usefixtures("credentials")
def test_publish_example_twine_fails_with_already_exists(
    snapshot, scenario_dist: Path, mock_twine: MockBinary
):
    mock_twine.set_error(
        """
Uploading distributions to https://test.pypi.org/legacy/
Uploading example_9e723676_a-1.0.0.tar.gz
100% ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━ 3.1/3.1 kB • 00:00 • ?
WARNING  Error during upload. Retry with the --verbose option for more details.
ERROR    HTTPError: 400 Bad Request from https://test.pypi.org/legacy/
         File already exists. See https://test.pypi.org/help/#file-name-reuse for more information.
        """
    )

    assert (
        snapshot_command(
            ["publish", scenario_dist, "-v", "--workers", "1"],
            extra_filters=[(re.escape(str(scenario_dist.resolve())), "[DISTDIR]")],
        )
        == snapshot
    )


@pytest.mark.usefixtures("tmpenviron")
def test_publish_example_no_credentials(
    snapshot,
    scenario_dist: Path,
    mock_twine: MockBinary,
):
    # Ensure these do not exist
    if "PACKSE_PUBLISH_PASSWORD" in os.environ:
        os.environ.pop("PACKSE_PUBLISH_PASSWORD")
    if "TWINE_PASSWORD" in os.environ:
        os.environ.pop("TWINE_PASSWORD")

    mock_twine.set_error(
        """
        Should not be reached!
        """
    )

    assert (
        snapshot_command(
            ["publish", scenario_dist, "-v", "--workers", "1"],
            extra_filters=[(re.escape(str(scenario_dist.resolve())), "[DISTDIR]")],
        )
        == snapshot
    )


@pytest.mark.usefixtures("credentials")
def test_publish_example_no_username_defaults_to_token(
    snapshot,
    scenario_dist: Path,
    mock_twine: MockBinary,
):
    # Ensure these do not exist
    if "PACKSE_PUBLISH_USERNAME" in os.environ:
        os.environ.pop("PACKSE_PUBLISH_USERNAME")
    if "TWINE_USERNAME" in os.environ:
        os.environ.pop("TWINE_USERNAME")

    assert (
        snapshot_command(
            ["publish", scenario_dist, "-v", "--workers", "1"],
            extra_filters=[(re.escape(str(scenario_dist.resolve())), "[DISTDIR]")],
        )
        == snapshot
    )



import fnmatch
import logging
import re
from pathlib import Path
from typing import Any

import chevron_blue
import msgspec

from packse.templates import __templates_path__

TEMPLATE_CONFIG = "template.toml"

logger = logging.getLogger(__name__)


class TemplateConfig(msgspec.Struct):
    version: int
    build_base: list[str]
    build_sdist: list[str]
    build_wheel: list[str]


def load_template_config(template_name: str) -> TemplateConfig:
    template_path = __templates_path__ / template_name
    config = template_path / TEMPLATE_CONFIG
    if not config.exists():
        raise RuntimeError(f"Template {template_name} missing config file.")
    return msgspec.toml.decode(config.read_text(), type=TemplateConfig)


def create_from_template(
    destination: Path, template_name: str, variables: dict[str, Any]
) -> Path:
    template_path = __templates_path__ / template_name
    first_root = None
    ignore_files = ["*.pyc"]
    for root, _, files in template_path.walk():
        for ignore in ignore_files:
            files = [file for file in files if not fnmatch.fnmatch(file, ignore)]

        for loop_root, scope in parse_loop(root, variables):
            # Determine the new directory path in the destination
            new_root = destination / Path(
                chevron_blue.render(str(loop_root), scope, no_escape=True)
            ).relative_to(template_path)

            if new_root == destination:
                continue

            if not first_root:
                first_root = new_root

            # Create the new directory
            logger.debug("Creating %s", new_root.relative_to(destination))
            new_root.mkdir()

            for file in files:
                file_path = root / file

                # Determine the new file path
                new_file_path = new_root / chevron_blue.render(
                    file, scope, no_escape=True
                )
                logger.debug("Creating %s", new_file_path.relative_to(destination))

                new_file_path.write_text(
                    chevron_blue.render(file_path.read_text(), scope, no_escape=True)
                )

    return first_root


def parse_loop(path: Path, variables: dict[str, Any]):
    # This implementation is pretty dubious and certain to fail on edge-cases
    scope = variables
    scopes = []
    matches = re.findall(r"\[\[(.*?)\]\]", str(path))
    if not matches:
        return [(path, variables)]

    scopes = variables[matches[0].strip()]
    for match in matches[1:]:
        match: str = match.strip()
        scopes = [scope[match] for scope in scopes if match in scope]

    results = []
    for scope in scopes:
        for inner in scope:
            results.append(
                (
                    Path(re.sub(r"\[\[(.*?)\]\]", "", str(path))),
                    inner,
                )
            )
    return results



import pytest
from packse import __development_base_path__

from .common import snapshot_command


def test_view_no_target(snapshot):
    assert (
        snapshot_command(["view"], snapshot_stderr=False, snapshot_stdout=False)
        == snapshot
    )


def test_view_target_does_not_exist(snapshot):
    assert snapshot_command(["view", "foo"]) == snapshot


@pytest.mark.usefixtures("tmpcwd")
def test_view_one_target_does_not_exist(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["view", str(target), "foo"]) == snapshot


def test_view_invalid_target(snapshot, tmpcwd):
    target = tmpcwd / "test.json"
    target.touch()
    assert snapshot_command(["view", str(target)]) == snapshot


def test_view_example(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["view", str(target)]) == snapshot


def test_view_example_short_names(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["view", str(target), "--short-names"]) == snapshot


def test_view_example_name(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["view", str(target), "--name", "example"]) == snapshot



"""
Utilities for testing
"""

import base64
import hashlib
import os
import re
import signal
import subprocess
import sys
from contextlib import contextmanager
from itertools import islice
from pathlib import Path

import pytest
from packse import __development_base_path__

MAX_FILE_LENGTH = 500
"""Maximum number of characters to store per file in filesystem snapshots. If exceeded, a hash of the contents will be used instead."""


def snapshot_command(
    command: list[str],
    working_directory: Path | None = None,
    snapshot_filesystem: bool = False,
    snapshot_stderr: bool = True,
    snapshot_stdout: bool = True,
    extra_filters: list[tuple[str, str]] | None = None,
    interrupt_after: float = None,
    env: dict[str, str] = None,
) -> dict:
    env = env or {}
    # By default, filter out absolute references to the working directory
    filters = [
        (re.escape(str(Path(sys.executable).parent)), "[PYTHON_BINDIR]"),
        (re.escape(str((working_directory or Path.cwd()).absolute())), "[PWD]"),
        (
            r"in (\d+\.)?\d+(ms|s)",
            "in [TIME]",
        ),
        (re.escape(str(__development_base_path__.absolute())), "[PROJECT_ROOT]"),
    ]
    if extra_filters:
        filters += extra_filters

    killed = False
    env = {**os.environ, **env}
    process = subprocess.Popen(
        ["packse"] + command,
        cwd=working_directory,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
    )
    try:
        stdout, stderr = process.communicate(timeout=interrupt_after)
    except subprocess.TimeoutExpired:
        process.send_signal(signal.SIGINT)
        stdout, stderr = process.communicate(timeout=2)
        killed = True

    result = {
        "exit_code": process.returncode if not killed else "<stopped>",
        "stdout": (
            apply_filters(stdout.decode(), filters)
            if snapshot_stdout
            else "<not included>"
        ),
        "stderr": (
            apply_filters(stderr.decode(), filters)
            if snapshot_stderr
            else "<not included>"
        ),
    }

    if snapshot_filesystem:
        result["filesystem"] = snapshot_filesystem_tree(working_directory)

    return result


@contextmanager
def tmpchdir(path: str):
    """
    Change current-working directories for the duration of the context
    """
    path = os.path.abspath(path)
    if os.path.isfile(path) or (not os.path.exists(path) and not path.endswith("/")):
        path = os.path.dirname(path)

    owd = os.getcwd()

    try:
        os.chdir(path)
        yield path
    finally:
        os.chdir(owd)


@pytest.fixture
def tmpcwd(tmp_path):
    with tmpchdir(tmp_path):
        yield tmp_path


@pytest.fixture
def tmpenviron():
    old_environ = dict(os.environ)

    yield os.environ

    os.environ.clear()
    os.environ.update(old_environ)


def snapshot_filesystem_tree(working_directory: Path | None = None) -> dict:
    working_directory = working_directory or Path.cwd()
    snapshot = {}
    snapshot["tree"] = tree(working_directory)

    for root, _, files in working_directory.walk():
        for file in files:
            file_path = root / file
            try:
                contents = file_path.read_text()
            except UnicodeDecodeError:
                contents = base64.b64encode(file_path.read_bytes())

            snapshot[str(file_path.relative_to(working_directory))] = (
                f"md5:{hash_contents(contents)}"
                if len(contents) > MAX_FILE_LENGTH
                else contents
            )

    return snapshot


def hash_contents(contents: bytes | str) -> str:
    if isinstance(contents, str):
        contents = contents.encode()
    return hashlib.md5(contents, usedforsecurity=False).hexdigest()


def apply_filters(content: str, filters: list[tuple[str, str]]) -> str:
    for pattern, replacement in filters:
        content = re.sub(pattern, replacement, content)
    return content


# Derived from https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python
def tree(
    dir_path: Path,
    level: int = -1,
    limit_to_directories: bool = False,
    length_limit: int = 1000,
):
    """Given a directory Path object print a visual tree structure"""

    space = "    "
    branch = "│   "
    tee = "├── "
    last = "└── "
    buffer = ""

    dir_path = Path(dir_path)  # accept string coerceable to Path
    files = 0
    directories = 0

    def inner(dir_path: Path, prefix: str = "", level=-1):
        nonlocal files, directories
        if not level:
            return  # 0, stop iterating
        if limit_to_directories:
            contents = [d for d in dir_path.iterdir() if d.is_dir()]
        else:
            contents = list(dir_path.iterdir())
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, sorted(contents)):
            if path.is_dir():
                yield prefix + pointer + path.name
                directories += 1
                extension = branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level=level - 1)
            elif not limit_to_directories:
                yield prefix + pointer + path.name
                files += 1

    buffer += dir_path.name + "\n"
    iterator = inner(dir_path, level=level)
    for line in islice(iterator, length_limit):
        buffer += line + "\n"
    if next(iterator, None):
        buffer += f"... length_limit, {length_limit}, reached, counted:\n"
    buffer += (
        f"\n{directories} directories" + (f", {files} files" if files else "") + "\n"
    )
    return buffer



import errno
import hashlib
import importlib
import logging
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager, nullcontext
from pathlib import Path
from typing import Generator

from packse import __development_base_path__
from packse.build import build
from packse.error import (
    RequiresExtra,
    ServeAddressInUse,
    ServeAlreadyRunning,
    ServeCommandError,
)
from packse.inspect import inspect
from packse.template import create_from_template

logger = logging.getLogger(__name__)


def index_up(
    host: str = "localhost",
    port: int = 3141,
    background: bool = False,
    dist_dir: Path | None = None,
    reset: bool = False,
    offline: bool = False,
):
    try:
        importlib.import_module("pypiserver")
    except ImportError:
        raise RequiresExtra("index commands", "index")

    if not dist_dir:
        dist_dir = Path.cwd() / "dist"
    else:
        dist_dir = Path(dist_dir).resolve()

    if background:
        if pid := get_server_pid():
            if is_running(pid):
                raise ServeAlreadyRunning()

    if reset:
        logger.info("Clearing existing server data...")
        if dist_dir.exists():
            shutil.rmtree(dist_dir)

    state_path = get_packse_state_path()

    logger.debug("Storing server metadata at %s", state_path)
    server_log_path = state_path / "server.log" if background else None
    index_url = f"http://{host}:{port}"
    index_note = "offline index" if offline else "local index with PyPI fallback"
    logger.info("Starting %s at %s", index_note, index_url)
    with start_index_server(
        dist_dir,
        host,
        port,
        server_log_path,
        offline=offline,
    ) as server_process:
        # Copy build dependencies over
        add_vendored_build_deps(dist_dir, offline)

        if background:
            logger.info("Running in background with pid %s", server_process.pid)
            logger.info("Stop index server with `packse index down`.")
            print(index_url)

        else:
            debug = logger.getEffectiveLevel() <= logging.DEBUG
            if not debug:
                logger.info("Hiding server output, use `-v` to stream server logs")

            logger.info("Ready! [Stop with Ctrl-C]")

            line = ""
            while True:
                line = server_process.stdout.readline().decode()
                if debug:
                    print(line, end="")


def index_down() -> bool:
    state_path = get_packse_state_path(create=False)

    if not state_path.exists():
        logger.warning("No server detected!")
        return False

    if pid := get_server_pid():
        if not is_running(pid):
            logger.info("Server looks shutdown already.")
            reset_pidfile()
            return

        logger.info("Stopping server with pid %s...", pid)
        try:
            os.kill(pid, signal.SIGINT)
        except ProcessLookupError:
            # Just move on if it's already dead
            pass

        # Wait for the server to exit for 10s
        limit = 10
        start_time = time.time()
        while time.time() - start_time < limit:
            if not is_running(pid):
                break
            time.sleep(0.1)
        else:
            # After the limit is reached, send a stronger signal...
            logger.info("Server has not stopped after %ds, sending kill...", limit)
            os.kill(pid, signal.SIGKILL)

            while is_running(pid):
                time.sleep(0.1)

        reset_pidfile()
        logger.info("Stopped server!")
        return True
    else:
        logger.warning("No server detected!")
        return False


@contextmanager
def start_index_server(
    dist_dir: Path,
    host: str,
    port: int,
    server_log_path: Path | None,
    offline: bool,
) -> Generator[subprocess.Popen, None, None]:
    server_url = f"http://{host}:{port}"
    dist_dir.mkdir(parents=True, exist_ok=True)

    command = [
        sys.executable,
        "-m",
        "pypiserver",
        "run",
        "--host",
        host,
        "--port",
        str(port),
        "-v",
        # Disable auth
        "-P",
        ".",
        "-a",
        ".",
        str(dist_dir),
    ]

    if offline:
        command.append("--disable-fallback")

    logger.debug("Running: %s", " ".join(command))

    server_process = subprocess.Popen(
        command,
        stderr=subprocess.STDOUT,
        stdout=server_log_path.open("wb") if server_log_path else subprocess.PIPE,
    )

    try:
        # Wait for server to be ready
        with (
            server_log_path.open("rb")
            if server_log_path
            else nullcontext(server_process.stdout) as server_output
        ):
            line = ""
            lines = []

            # Watch for startup
            # TODO: Use a healthceck instead?
            while "Serving on" not in line and "Listening on" not in line:
                try:
                    server_process.wait(0)
                except subprocess.TimeoutExpired:
                    pass
                else:
                    # The process stopped already
                    if (
                        logger.getEffectiveLevel() > logging.DEBUG
                        and server_output.seekable()
                    ):
                        server_output.seek(0)
                    err_output = (
                        "\n".join(lines) + server_output.read().decode().strip()
                    )
                    raise ServeCommandError("Failed to start server!", err_output)

                line = server_output.readline().decode()
                if logger.getEffectiveLevel() <= logging.DEBUG:
                    print(line, end="")
                elif not server_output.seekable():
                    # Track the lines if we'll never be able to get them back
                    lines.append(line)
                if "Address already in use" in line:
                    raise ServeAddressInUse(server_url)

        # Server started!
        write_server_pid(server_process.pid)

        if server_log_path:
            logger.info("Writing server logs to %s", server_log_path)

        try:
            display_path = f"./{dist_dir.relative_to(Path.cwd())}"
        except Exception:
            display_path = dist_dir

        logger.info(f"Serving distributions from {display_path}")

        yield server_process

    except BaseException as exc:
        server_process.terminate()

        stdout, _ = server_process.communicate(timeout=1)
        if logger.getEffectiveLevel() <= logging.DEBUG:
            print(stdout.decode())

        server_process.kill()

        stdout, _ = server_process.communicate(timeout=1)
        if logger.getEffectiveLevel() <= logging.DEBUG:
            print(stdout.decode())

        # Do not reset the pidfile when a server was already running!
        if not isinstance(exc, ServeAddressInUse):
            reset_pidfile()

        raise


def add_vendored_build_deps(dist_dir: Path, offline: bool):
    build_directory = __development_base_path__ / "vendor" / "build"
    if build_directory.exists():
        logger.debug("Copying vendored build dependencies...")
        shutil.copytree(build_directory, dist_dir / "build", dirs_exist_ok=True)
    else:
        logger.warning("No vendored build dependencies found at %s" % build_directory)
        if offline:
            logger.error(
                "Server has no fallback PyPI access and source distribution builds are likely to fail"
            )


def get_packse_state_path(create: bool = True) -> Path:
    path = Path(os.environ.get("PACKSE_STATE_PATH") or (Path.home() / ".packse"))
    if create:
        path.mkdir(exist_ok=True)
    return path


def get_pidfile_path() -> Path:
    return get_packse_state_path() / "server.pid"


def get_server_pid() -> int | None:
    pidfile = get_pidfile_path()
    if not pidfile.exists():
        return None
    else:
        logger.debug("Found existing pid at %s", pidfile)
        return int(pidfile.read_text())


def write_server_pid(pid: int) -> None:
    pidfile = get_pidfile_path()
    pidfile.write_text(str(pid))


def reset_pidfile():
    pidfile = get_pidfile_path()
    pidfile.unlink(missing_ok=True)


def is_running(pid):
    try:
        os.kill(pid, 0)
    except OSError as err:
        # ESRCH: pid not exist
        if err.errno == errno.ESRCH:
            return False
    return True


def sha256_file(path: Path):
    h = hashlib.sha256()

    with open(path, "rb") as file:
        while True:
            # Reading is buffered, so we can read smaller chunks.
            chunk = file.read(h.block_size)
            if not chunk:
                break
            h.update(chunk)

    return h.hexdigest()


def build_index(
    targets: list[Path],
    no_hash: bool,
    short_names: bool,
    dist_dir: Path | None,
):
    out_path = Path("./index")
    if out_path.exists():
        shutil.rmtree(out_path)

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        if not dist_dir:
            logger.info("Building distributions...")
            build(
                targets,
                rm_destination=True,
                skip_root=False,
                short_names=short_names,
                no_hash=no_hash,
                dist_dir=tmpdir / "dist",
                build_dir=tmpdir / "build",
            )
            dist_dir = tmpdir / "dist"

        out_path.mkdir()
        (out_path / "files").mkdir()

        variables = inspect(targets, short_names=short_names, no_hash=no_hash)
        for scenario in variables["scenarios"]:
            for package in scenario["packages"]:
                # Create a new distributions section
                package["dists"] = []
                for version in package["versions"]:
                    for file in Path(
                        dist_dir
                        / (
                            scenario["name"]
                            if no_hash
                            else f"{scenario['name']}-{scenario['version']}"
                        )
                    ).iterdir():
                        if package["name"].replace("-", "_") + "-" + version[
                            "version"
                        ] not in str(file):
                            continue

                        # Include all the version information to start
                        dist = version.copy()
                        # Then add a sha256
                        dist["sha256"] = sha256_file(file)
                        dist["file"] = file.name
                        package["dists"].append(dist)
                        logger.info(
                            "Found distribution %s",
                            file.name,
                        )
                        file.rename(out_path / "files" / file.name)

        logger.info("Populating template...")
        create_from_template(
            out_path,
            "index",
            variables=variables,
        )

    logger.info("Built static index at ./%s", out_path)



import os
import subprocess

from packse.index import write_server_pid

from .common import snapshot_command

FILTERS = [
    (r"pid \d+", "pid [PID]"),
    (r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d,\d\d\d", "[TIMESTAMP]"),
    ("uuid: .*", "uuid: [UUID]"),
]


def test_index_up_background(snapshot, tmpcwd, tmpenviron):
    tmpenviron["HOME"] = str(tmpcwd)

    try:
        assert (
            snapshot_command(
                ["index", "up", "--bg"],
                extra_filters=FILTERS,
            )
            == snapshot
        )
    finally:
        subprocess.call(["packse", "index", "down"], env=tmpenviron)


def test_index_up_foreground(snapshot, tmpcwd, tmpenviron):
    tmpenviron["HOME"] = str(tmpcwd)

    assert (
        snapshot_command(
            ["index", "up"],
            extra_filters=FILTERS,
            # Send a keyboard interrupt after a bit — use a longer delay for slow CI machines
            interrupt_after=15 if os.environ.get("CI") else 5,
        )
        == snapshot
    )


def test_index_down(snapshot, tmpcwd, tmpenviron):
    tmpenviron["HOME"] = str(tmpcwd)
    subprocess.check_call(["packse", "index", "up", "--bg"])
    assert (
        snapshot_command(
            ["index", "down"],
            extra_filters=FILTERS,
        )
        == snapshot
    )


def test_index_down_no_server_found(snapshot, tmpcwd):
    assert (
        snapshot_command(
            ["index", "down"],
            extra_filters=FILTERS,
            env={"PACKSE_STATE_PATH": str(tmpcwd)},
        )
        == snapshot
    )


def test_index_down_server_stopped(snapshot, tmpcwd, tmpenviron):
    tmpenviron["PACKSE_STATE_PATH"] = str(tmpcwd)
    write_server_pid(99999)

    assert (
        snapshot_command(
            ["index", "down"],
            extra_filters=FILTERS,
            env={"PACKSE_STATE_PATH": str(tmpcwd)},
        )
        == snapshot
    )



"""
Publish package distributions.
"""

import logging
import os
import subprocess
import sys
import textwrap
import time
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait as wait_for_futures
from pathlib import Path

from packse.error import (
    InvalidPublishTarget,
    PublishAlreadyExists,
    PublishConnectionError,
    PublishError,
    PublishNoCredentials,
    PublishRateLimit,
    PublishToolError,
)

logger = logging.getLogger(__name__)


# Rate limits are defined at https://github.com/pypi/warehouse/blob/d858a996fe543131931955d8e8cc96b8aa7261a1/warehouse/config.py#L352-L369
# At the time of writing, rate limits enforce a maximum of 20 new projects / hour.
# Generally, retries aren't going to help.
MAX_ATTEMPTS = 3
RETRY_TIME = 1  # Start with one second
RETRY_BACKOFF_FACTOR = 1.5


def publish(
    targets: list[Path],
    index_url: str,
    skip_existing: bool,
    dry_run: bool,
    workers: int,
    anonymous: bool,
):
    start_time = time.time()

    if not anonymous and not (
        "TWINE_PASSWORD" in os.environ or "PACKSE_PUBLISH_PASSWORD" in os.environ
    ):
        raise PublishNoCredentials()

    # Skip yanked markers
    targets = [target for target in targets if not str(target).endswith(".yanked")]

    for target in targets:
        if not target.is_dir():
            raise InvalidPublishTarget(target, reason="Not a directory.")

    s = "" if len(targets) == 1 else "s"
    logger.info("Publishing %s target%s to %s...", len(targets), s, index_url)
    for target in sorted(targets):
        logger.info("Publishing '%s'...", target.name)

    # Publish each directory
    with ThreadPoolExecutor(
        thread_name_prefix="packse-scenario-", max_workers=workers
    ) as executor:
        futures = [
            executor.submit(
                publish_package_distributions,
                target,
                index_url,
                skip_existing=skip_existing,
                dry_run=dry_run,
                anonymous=anonymous,
            )
            for target in targets
        ]

        wait_for_futures(futures)

    # All the futures are done since the executor context has exited; check if any failed
    incomplete = 0
    for future in futures:
        if (exc := future.exception()) is None:
            continue
        incomplete += 1
        print(f"{exc}.", file=sys.stderr)

    if incomplete:
        s = "" if targets == 1 else "s"
        raise PublishError(f"Failed to publish {incomplete}/{len(targets)} target{s}")

    total_files = 0
    for name, files in sorted(future.result() for future in futures):
        print(name)
        total_files += files

    for target in targets:
        yankfile = target.with_suffix(".yanked")
        if yankfile.exists():
            print()
            print(f"{target.name} has versions that must be yanked:")
            for package_version in yankfile.read_text().splitlines():
                package, version = package_version.rsplit("-", 1)
                url = (
                    index_url.replace("/legacy/", "/")
                    + f"manage/project/{package}/release/{version}/#yank_version-modal"
                )
                print(f"\t{package_version}: {url}")

    logger.info(
        "Published %s targets (%s new files) in %.2fs",
        len(futures),
        total_files,
        time.time() - start_time,
    )


def publish_package_distributions(
    target: Path,
    index_url: str,
    skip_existing: bool,
    dry_run: bool,
    anonymous: bool,
) -> tuple[str, int]:
    """
    Publish a directory of package distribution files.
    """
    files = 0
    for distfile in sorted(target.iterdir()):
        if publish_package_distribution_with_retries(
            distfile,
            index_url,
            skip_existing=skip_existing,
            dry_run=dry_run,
            anonymous=anonymous,
            max_attempts=MAX_ATTEMPTS,
        ):
            files += 1

    return (target.name, files)


def publish_package_distribution_with_retries(
    target: Path,
    index_url: str,
    skip_existing: bool,
    dry_run: bool,
    anonymous: bool,
    max_attempts: int,
) -> bool:
    """
    Publish a package distribution file with retries and error handling.
    """
    attempts = 0
    retry_time = RETRY_TIME
    while attempts < max_attempts:
        retry_time = retry_time * RETRY_BACKOFF_FACTOR
        attempts += 1
        try:
            publish_package_distribution(
                target, index_url, anonymous=anonymous, dry_run=dry_run
            )
        except PublishAlreadyExists:
            if not skip_existing:
                raise
            logger.info("Skipped '%s': already published.", target.name)
            return False
        except PublishConnectionError:
            if attempts >= max_attempts:
                raise
            logger.warning(
                "Encountered connection error publishing '%s', retrying in %ds",
                target.name,
                retry_time,
            )
            time.sleep(retry_time)
        else:
            logger.info("Published '%s'", target.name)
            break

    return True


def publish_package_distribution(
    target: Path, index_url: str, anonymous: bool, dry_run: bool
) -> None:
    """
    Publish a package distribution file.
    """
    command = ["twine", "upload", "--repository-url", index_url, str(target.resolve())]
    if dry_run:
        print("Would execute: " + " ".join(command))
        return

    start_time = time.time()
    try:
        env = os.environ.copy()
        # Ensure twine does not prompt for credentials
        env["TWINE_NON_INTERACTIVE"] = "1"

        # Pass the publish username through to twine
        if publish_username := os.environ.get("PACKSE_PUBLISH_USERNAME"):
            env["TWINE_USERNAME"] = publish_username

        # Configure the username for tokens by default
        env.setdefault("TWINE_USERNAME", "__token__")

        # Pass the publish token through to twine
        if publish_token := os.environ.get("PACKSE_PUBLISH_PASSWORD"):
            env["TWINE_PASSWORD"] = publish_token

        # Twine requires a username and password even if we don't want to use them
        if anonymous:
            env["TWINE_USERNAME"] = "ANON"
            env["TWINE_PASSWORD"] = "ANON"

        output = subprocess.check_output(
            command, stderr=subprocess.STDOUT, env=env, timeout=60
        )
    except subprocess.TimeoutExpired:
        raise PublishError(f"Publish of {target.name} timed out.")
    except subprocess.CalledProcessError as exc:
        output = exc.output.decode()
        if "File already exists" in output or "409 Conflict" in output:
            raise PublishAlreadyExists(target.name)
        if "HTTPError: 429 Too Many Requests" in output:
            raise PublishRateLimit(target.name)
        if "ConnectionError" in output:
            raise PublishConnectionError(target.name)
        raise PublishToolError(
            f"Publishing {target.name} with twine failed",
            output,
        )
    else:
        logs = (
            (":\n\n" + textwrap.indent(output.decode(), " " * 4))
            if logger.getEffectiveLevel() <= logging.DEBUG
            else ""
        )
        logger.debug(
            "Published %s in %.2fs%s",
            target.name,
            time.time() - start_time,
            logs,
        )



from .common import snapshot_command


def test_fetch(snapshot, tmpcwd):
    assert (
        # Include a ref for reproducibility
        snapshot_command(
            ["fetch", "--ref", "df20b898fdf1fd242cc19acc2a3148d72aa4d89f"],
            snapshot_filesystem=True,
            working_directory=tmpcwd,
        )
        == snapshot
    )


def test_fetch_dest(snapshot, tmpcwd):
    assert (
        snapshot_command(
            [
                "fetch",
                "--dest",
                "foo",
                "--ref",
                "df20b898fdf1fd242cc19acc2a3148d72aa4d89f",
            ],
            snapshot_filesystem=True,
            working_directory=tmpcwd,
        )
        == snapshot
    )


def test_fetch_dest_exists(snapshot, tmpcwd):
    (tmpcwd / "foo").mkdir()
    assert (
        snapshot_command(
            [
                "fetch",
                "--dest",
                "foo",
                "--ref",
                "df20b898fdf1fd242cc19acc2a3148d72aa4d89f",
            ],
            snapshot_filesystem=True,
            working_directory=tmpcwd,
        )
        == snapshot
    )


def test_fetch_dest_exists_force(snapshot, tmpcwd):
    (tmpcwd / "foo").mkdir()
    assert (
        snapshot_command(
            [
                "fetch",
                "--dest",
                "foo",
                "--force",
                "--ref",
                "df20b898fdf1fd242cc19acc2a3148d72aa4d89f",
            ],
            snapshot_filesystem=True,
            working_directory=tmpcwd,
        )
        == snapshot
    )


def test_fetch_tag(snapshot, tmpcwd):
    assert (
        snapshot_command(
            ["fetch", "--ref", "0.1.0"],
            snapshot_filesystem=True,
            working_directory=tmpcwd,
        )
        == snapshot
    )



import os
from pathlib import Path

import pytest
from packse import __development_base_path__
from packse.scenario import load_scenario, scenario_hash

from .common import snapshot_command


def test_build_no_target(snapshot):
    assert snapshot_command(["build"]) == snapshot


def test_build_target_does_not_exist(snapshot):
    assert snapshot_command(["build", "foo"]) == snapshot


def test_build_invalid_target(snapshot, tmpcwd):
    target = tmpcwd / "test.json"
    target.touch()
    assert snapshot_command(["build", str(target)]) == snapshot


@pytest.mark.usefixtures("tmpcwd")
def test_build_example(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert (
        snapshot_command(
            ["build", str(target)], snapshot_filesystem=True, snapshot_stderr=False
        )
        == snapshot
    )


@pytest.mark.usefixtures("tmpcwd")
def test_build_example_short_names(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert (
        snapshot_command(
            ["build", str(target), "--short-names"],
            snapshot_filesystem=True,
            snapshot_stderr=False,
        )
        == snapshot
    )


@pytest.mark.usefixtures("tmpcwd")
@pytest.mark.usefixtures("tmpenviron")
def test_build_example_with_seed(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    os.environ["PACKSE_VERSION_SEED"] = "foo"
    assert (
        snapshot_command(
            ["build", str(target)], snapshot_filesystem=True, snapshot_stderr=False
        )
        == snapshot
    )


@pytest.mark.usefixtures("tmpcwd")
def test_build_example_already_exists(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    scenario = load_scenario(target)
    name = f"{scenario.name}-{scenario_hash(scenario)}"
    (Path.cwd() / "build" / name).mkdir(parents=True)
    assert snapshot_command(["build", target], snapshot_filesystem=True) == snapshot


@pytest.mark.usefixtures("tmpcwd")
def test_build_example_already_exists_with_rm_flag(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    (Path(".") / "build").mkdir()
    assert (
        snapshot_command(["build", target, "--rm"], snapshot_stderr=False) == snapshot
    )



import pytest
from packse import __development_base_path__

from .common import snapshot_command


def test_list_no_target_finds_all_valid_scenarios(snapshot):
    assert (
        snapshot_command(["list"], snapshot_stdout=False, snapshot_stderr=False)
        == snapshot
    )


def test_list_no_hash(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["list", str(target), "--no-hash"]) == snapshot


def test_list_short_names(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["list", str(target), "--short-names"]) == snapshot


def test_list_no_sources(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["list", str(target), "--no-sources"]) == snapshot


def test_list_target_does_not_exist(snapshot):
    assert snapshot_command(["list", "foo"]) == snapshot


@pytest.mark.usefixtures("tmpcwd")
def test_list_one_target_does_not_exist(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["list", str(target), "foo"]) == snapshot


def test_list_invalid_target(snapshot, tmpcwd):
    bad_target = tmpcwd / "test.json"
    bad_target.touch()
    good_target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["list", str(bad_target), str(good_target)]) == snapshot


def test_list_invalid_target_skip_invalid(snapshot, tmpcwd):
    bad_target = tmpcwd / "test.json"
    bad_target.touch()
    good_target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert (
        snapshot_command(["list", str(bad_target), str(good_target), "--skip-invalid"])
        == snapshot
    )



import pytest
from packse import __development_base_path__

from .common import snapshot_command


def test_inspect_no_target_finds_all_valid_scenarios(snapshot):
    assert (
        # Just assert it succeeds, we don't want to include all these targets in the snap
        snapshot_command(["inspect"], snapshot_stdout=False, snapshot_stderr=False)
        == snapshot
    )


def test_inspect_target_does_not_exist(snapshot):
    assert snapshot_command(["inspect", "foo"]) == snapshot


@pytest.mark.usefixtures("tmpcwd")
def test_inspect_one_target_does_not_exist(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["inspect", str(target), "foo"]) == snapshot


def test_inspect_invalid_target(snapshot, tmpcwd):
    bad_target = tmpcwd / "test.json"
    bad_target.touch()
    good_target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["inspect", str(bad_target), str(good_target)]) == snapshot


def test_inspect_invalid_target_skip_invalid(snapshot, tmpcwd):
    bad_target = tmpcwd / "test.json"
    bad_target.touch()
    good_target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert (
        snapshot_command(
            ["inspect", str(bad_target), str(good_target), "--skip-invalid"]
        )
        == snapshot
    )


def test_inspect_json(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.json"
    assert snapshot_command(["inspect", str(target)]) == snapshot


def test_inspect_toml(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.toml"
    assert snapshot_command(["inspect", str(target)]) == snapshot


def test_inspect_yaml(snapshot):
    target = __development_base_path__ / "scenarios" / "examples" / "example.yaml"
    assert snapshot_command(["inspect", str(target)]) == snapshot



"""
Get details for all scenarios.
"""

import logging
from pathlib import Path
from typing import cast

from packse.error import FileNotFound, InvalidScenario
from packse.scenario import (
    Requirement,
    Scenario,
    find_scenario_files,
    load_scenarios,
    scenario_hash,
)
from packse.view import dependency_tree

logger = logging.getLogger(__name__)


def inspect(
    targets: list[Path],
    skip_invalid: bool = False,
    short_names: bool = False,
    no_hash: bool = False,
):
    scenarios_by_path: dict[Path, list[Scenario]] = {}

    # Validate and collect all targets first
    for target in sorted(targets):
        if not target.exists():
            raise FileNotFound(target)

        if target.is_dir():
            for target in find_scenario_files(target):
                try:
                    logger.debug("Loading %s", target)
                    scenarios_by_path[target] = load_scenarios(target)
                except Exception as exc:
                    invalid = InvalidScenario(target, reason=str(exc))
                    if skip_invalid:
                        print(f"Skipping file: {invalid}")
                    else:
                        raise invalid
        else:
            try:
                logger.debug("Loading %s", target)
                scenarios_by_path[target] = load_scenarios(target)
            except Exception as exc:
                invalid = InvalidScenario(target, reason=str(exc))
                if skip_invalid:
                    print(f"Skipping file: {invalid}")
                else:
                    raise invalid

    # Collect a JSON-compatible representation for each scenario
    result = {"scenarios": []}
    for source, scenarios in scenarios_by_path.items():
        for scenario in scenarios:
            scenario = cast(Scenario, scenario)

            raw = scenario.dict()
            raw["source"] = str(source)
            raw["version"] = scenario_hash(scenario)
            raw["tree"] = dependency_tree(scenario).splitlines()
            result["scenarios"].append(raw)

            raw["resolver_options"]["no_build"] = [
                str(
                    Requirement(name).with_unique_name(
                        scenario,
                        raw["version"],
                        short_names,
                        no_hash=no_hash,
                    )
                )
                for name in raw["resolver_options"]["no_build"]
            ]
            raw["resolver_options"]["no_binary"] = [
                str(
                    Requirement(name).with_unique_name(
                        scenario,
                        raw["version"],
                        short_names,
                        no_hash=no_hash,
                    )
                )
                for name in raw["resolver_options"]["no_binary"]
            ]

            # Convert dictionaries to lists for easier templating
            raw["packages"] = [
                {
                    "name": str(
                        Requirement(name).with_unique_name(
                            scenario,
                            raw["version"],
                            short_names,
                            no_hash=no_hash,
                        )
                    ),
                    "versions": [
                        {
                            **package_metadata.dict(),
                            "version": version,
                            "requires": [
                                str(
                                    requirement.with_unique_name(
                                        scenario,
                                        raw["version"],
                                        short_names,
                                        no_hash=no_hash,
                                    )
                                )
                                for requirement in package_metadata.requires
                            ],
                            "extras": [
                                {
                                    "name": extra,
                                    "requires": [
                                        str(
                                            requirement.with_unique_name(
                                                scenario,
                                                raw["version"],
                                                short_names,
                                                no_hash=no_hash,
                                            )
                                        )
                                        for requirement in extra_requires
                                    ],
                                }
                                for extra, extra_requires in package_metadata.extras.items()
                            ],
                        }
                        for version, package_metadata in package.versions.items()
                    ],
                }
                for name, package in scenario.packages.items()
            ]
            raw["expected"]["packages"] = [
                {
                    "name": str(
                        Requirement(name).with_unique_name(
                            scenario,
                            raw["version"],
                            short_names,
                            no_hash=no_hash,
                        )
                    ),
                    "version": version,
                }
                for name, version in scenario.expected.packages.items()
            ]
            raw["root"]["requires"] = [
                {
                    "requirement": str(requirement),
                    "name": requirement.name,
                    "module_name": requirement.name.replace("-", "_"),
                }
                for requirement in (
                    requirement.with_unique_name(
                        scenario,
                        raw["version"],
                        short_names,
                        no_hash=no_hash,
                    )
                    for requirement in scenario.root.requires
                )
            ]

            # Ensure a module name is available for testing import of packages
            for package in raw["expected"]["packages"]:
                package["module_name"] = package["name"].replace("-", "_")

            raw["module_name"] = raw["name"].replace("-", "_")

    return result



import textwrap
from pathlib import Path


class PackseError(Exception):
    """Base type for all Packse exceptions"""


class UserError(PackseError):
    """It's the user's fault :)"""


class RequiresExtra(UserError):
    def __init__(self, feature: str, extra: str) -> None:
        message = f"Using {feature} requires installation with extra {extra!r}"
        super().__init__(message)


class DestinationAlreadyExists(UserError):
    def __init__(self, destination: Path) -> None:
        message = f"Destination directory '{destination}' already exists"
        super().__init__(message)


class InvalidScenario(UserError):
    def __init__(self, path: Path, reason: str) -> None:
        message = f"File at '{path}' is not a valid scenario: {reason}"
        super().__init__(message)


class InvalidPackageVersion(UserError):
    """Version is not PEP compliant"""

    def __init__(self, version: str) -> None:
        message = f"Version {version!r} is not valid."
        super().__init__(message)


class FileNotFound(UserError):
    def __init__(self, path: Path) -> None:
        message = f"File '{path}' not found"
        super().__init__(message)


class BuildError(PackseError):
    def __init__(self, message, output) -> None:
        message = message + ":\n" + textwrap.indent(output, " " * 4)
        super().__init__(message)


class PublishError(PackseError):
    pass


class PublishToolError(PublishError):
    def __init__(self, message, output) -> None:
        message = message + ":\n" + textwrap.indent(output, " " * 4)
        super().__init__(message)


class PublishNoCredentials(PublishError):
    def __init__(self) -> None:
        message = (
            "No credentials found for publish!\n"
            "Provide an API token via `PACKSE_PUBLISH_PASSWORD` or disable authentication with `--anonymous`"
        )
        super().__init__(message)


class PublishAlreadyExists(PublishError):
    def __init__(self, package: str) -> None:
        message = f"Publish for '{package}' already exists"
        super().__init__(message)


class PublishRateLimit(PublishError):
    def __init__(self, package: str) -> None:
        message = f"Publish of '{package}' failed due to rate limits"
        super().__init__(message)


class PublishConnectionError(PublishError):
    def __init__(self, package: str) -> None:
        message = f"Publish of '{package}' failed due to a connection error"
        super().__init__(message)


class InvalidPublishTarget(UserError):
    def __init__(self, path: Path, reason: str) -> None:
        message = f"Publish target at '{path}' is not a valid: {reason}"
        super().__init__(message)


class ServeError(PackseError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class ServeThreadError(ServeError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class ServeCommandError(ServeError):
    def __init__(self, message: str, stderr: str) -> None:
        if stderr.strip():
            message += f":\n\n{stderr}"
        super().__init__(message)


class ServeAddressInUse(ServeError):
    def __init__(self, url: str) -> None:
        message = f"Address '{url}' already in use"
        super().__init__(message)


class ServeAlreadyRunning(ServeError):
    def __init__(self) -> None:
        message = "Server is already running"
        super().__init__(message)



import argparse
import json
import logging
import sys
from pathlib import Path
from subprocess import CalledProcessError

from packse.build import build, build_package
from packse.error import (
    BuildError,
    DestinationAlreadyExists,
    PublishError,
    ServeCommandError,
    ServeError,
    UserError,
)
from packse.fetch import fetch
from packse.index import build_index, index_down, index_up
from packse.inspect import inspect
from packse.list import list
from packse.publish import publish
from packse.scenario import find_scenario_files
from packse.serve import serve
from packse.view import view


def entrypoint():
    parser = get_parser()
    args = parser.parse_args()
    if not hasattr(args, "call"):
        parser.print_help()
        return None

    if args.quiet:
        log_level = logging.CRITICAL
    elif args.verbose:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    logging.basicConfig(level=log_level, format="%(message)s")

    # Call the implementation wrapper e.g. `_call_build`
    try:
        args.call(args)
    except DestinationAlreadyExists as exc:
        print(f"{exc}. Pass `--force` to allow removal.", file=sys.stderr)
        exit(1)
    except UserError as exc:
        print(f"{exc}.", file=sys.stderr)
        exit(1)
    except BuildError as exc:
        print(f"{exc}.", file=sys.stderr)
        exit(1)
    except PublishError as exc:
        print(f"{exc}.", file=sys.stderr)
        exit(1)
    except ServeCommandError as exc:
        print(f"{exc}", file=sys.stderr)
        exit(1)
    except ServeError as exc:
        print(f"{exc}.", file=sys.stderr)
        exit(1)
    except CalledProcessError as exc:
        print(
            f"Error running command: {', '.join(exc.cmd)!r} (exit code {exc.returncode})",
            file=sys.stderr,
        )
        print(exc.output.decode(), file=sys.stderr)
        exit(2)
    except KeyboardInterrupt:
        print("Interrupted!", file=sys.stderr)
        exit(1)


def _call_build(args):
    build(
        args.targets,
        rm_destination=args.force,
        short_names=args.short_names,
        no_hash=args.no_hash,
        skip_root=args.skip_root,
    )


def _call_build_package(args):
    build_package(
        args.name,
        args.version,
        args.no_wheel,
        args.no_sdist,
        args.requires_python,
        args.wheel_tag,
        args.force,
    )


def _call_view(args):
    view(args.targets, args.name, short_names=args.short_names)


def _call_serve(args):
    serve(
        args.targets,
        host=args.host,
        port=args.port,
        dist_dir=args.dist_dir,
        build_dir=args.build_dir,
        short_names=args.short_names,
        no_hash=args.no_hash,
        offline=args.offline,
    )


def _call_index_up(args):
    index_up(
        host=args.host,
        port=args.port,
        reset=args.reset,
        dist_dir=args.dist_dir,
        background=args.bg,
        offline=args.offline,
    )


def _call_index_down(args):
    success = index_down()
    if not success:
        exit(1)


def _call_index_build(args):
    build_index(
        args.targets,
        short_names=args.short_names,
        no_hash=args.no_hash,
        dist_dir=args.dist_dir,
    )


def _call_publish(args):
    publish(
        args.targets,
        index_url=args.index_url,
        dry_run=args.dry_run,
        skip_existing=args.skip_existing,
        anonymous=args.anonymous,
        workers=args.workers,
    )


def _call_fetch(args):
    fetch(
        args.dest,
        ref=args.ref,
        force=args.force,
    )


def _call_list(args):
    skip_invalid = args.skip_invalid
    if not args.targets:
        skip_invalid = True
        targets = find_scenario_files(Path.cwd())
    else:
        targets = []
        for target in args.targets:
            # Expand any directories to json files within
            if target.is_dir():
                targets.extend(find_scenario_files(target))
            else:
                targets.append(target)

    list(
        targets,
        args.no_hash,
        skip_invalid,
        args.no_sources,
        short_names=args.short_names,
    )


def _call_inspect(args):
    skip_invalid = args.skip_invalid
    if not args.targets:
        skip_invalid = True
        targets = find_scenario_files(Path.cwd())
    else:
        targets = []
        for target in args.targets:
            # Expand any directories to scenario files within
            if target.is_dir():
                targets.extend(find_scenario_files(target))
            else:
                targets.append(target)

    print(
        json.dumps(
            inspect(
                targets,
                skip_invalid,
                short_names=args.short_names,
                no_hash=args.no_hash,
            ),
            indent=2,
        )
    )


def _root_parser():
    parser = argparse.ArgumentParser(
        description="Utilities for working example packaging scenarios",
    )
    _add_shared_arguments(parser)
    return parser


def _add_build_parser(subparsers):
    parser = subparsers.add_parser("build", help="Build packages for a scenario")
    parser.set_defaults(call=_call_build)
    parser.add_argument(
        "targets",
        type=Path,
        nargs="+",
        help="The scenario to build",
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Overwrite existing builds",
    )
    parser.add_argument(
        "--short-names",
        action="store_true",
        help="Exclude scenario names from generated packages.",
    )
    parser.add_argument(
        "--no-hash",
        action="store_true",
        help="Exclude scenario version hashes from generated packages.",
    )
    parser.add_argument(
        "--skip-root",
        action="store_true",
        help='Do not build a "root" package for the scenario.',
    )
    _add_shared_arguments(parser)


def _add_build_package_parser(subparsers):
    parser = subparsers.add_parser("build-pkg", help="Build a single package")
    parser.set_defaults(call=_call_build_package)
    parser.add_argument(
        "name",
        type=str,
        help="The package name",
    )
    parser.add_argument(
        "version",
        type=str,
        help="The package version",
    )
    parser.add_argument(
        "-t",
        "--wheel-tag",
        type=str,
        help="The tags for wheels",
        action="append",
    )
    parser.add_argument(
        "--no-wheel",
        action="store_true",
        help="Disable building wheels",
    )
    parser.add_argument(
        "--requires-python",
        type=str,
        default=None,
        help="The package Python requirement",
    )
    parser.add_argument(
        "--no-sdist",
        action="store_true",
        help="Disable building source distributions",
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Replace existing builds",
    )
    _add_shared_arguments(parser)


def _add_publish_parser(subparsers):
    parser: argparse.ArgumentParser = subparsers.add_parser(
        "publish", help="Publish packages for a scenario"
    )
    parser.set_defaults(call=_call_publish)
    parser.add_argument(
        "targets",
        type=Path,
        nargs="+",
        help="The scenario distribution directory to publish",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Do not actually publish, just show twine commands that would be used.",
    )
    parser.add_argument(
        "--skip-existing",
        action="store_true",
        help="Skip existing distributions instead of failing.",
    )
    parser.add_argument(
        "--workers",
        type=int,
        default=8,
        help="Number of threads to use when publishing.",
    )
    parser.add_argument(
        "--index-url",
        type=str,
        default="https://test.pypi.org/legacy/",
        help="The URL of the package index to publish the packages to.",
    )
    parser.add_argument(
        "--anonymous", action="store_true", help="Upload without credentials."
    )
    _add_shared_arguments(parser)


def _add_view_parser(subparsers):
    parser = subparsers.add_parser("view", help="View a scenario")
    parser.set_defaults(call=_call_view)
    parser.add_argument(
        "targets",
        type=Path,
        nargs="+",
        help="The scenario file to view",
    )
    parser.add_argument(
        "--name",
        type=str,
        default=None,
        help="The scenario name to view",
    )
    parser.add_argument(
        "--short-names",
        action="store_true",
        help="Exclude scenario names from generated packages.",
    )
    _add_shared_arguments(parser)


def _add_serve_parser(subparsers):
    parser = subparsers.add_parser(
        "serve", help="Serve scenarios on a temporary local package index"
    )
    parser.set_defaults(call=_call_serve)
    parser.add_argument(
        "targets",
        type=Path,
        nargs="*",
        help="The scenarios to serve",
    )
    parser.add_argument(
        "--host",
        type=str,
        default="localhost",
        help="The host to bind the package index to.",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=3141,
        help="The port to bind the package index to.",
    )
    parser.add_argument(
        "--dist-dir",
        type=Path,
        default="./dist",
        help="The directory to store and serve builds from.",
    )
    parser.add_argument(
        "--build-dir",
        type=Path,
        default="./build",
        help="The directory to store intermediate build artifacts in.",
    )
    parser.add_argument(
        "--short-names",
        action="store_true",
        help="Exclude scenario names from generated packages.",
    )
    parser.add_argument(
        "--no-hash",
        action="store_true",
        help="Exclude scenario version hashes from generated packages.",
    )
    parser.add_argument(
        "--offline",
        action="store_true",
        help="Run the index servers without fallback access to the real PyPI.",
    )
    _add_shared_arguments(parser)


def _add_index_parser(subparsers):
    parser = subparsers.add_parser("index", help="Run a local package index")

    subparsers = parser.add_subparsers(title="commands")
    up = subparsers.add_parser(
        "up", help="Start a package index server in the background."
    )
    up.add_argument(
        "--host",
        type=str,
        default="localhost",
        help="The host to bind the package index to.",
    )
    up.add_argument(
        "--port",
        type=int,
        default=3141,
        help="The port to bind the package index to.",
    )
    up.add_argument(
        "--reset",
        action="store_true",
        help="Reset the server's state on start.",
    )
    up.add_argument(
        "--dist-dir",
        type=Path,
        default="./dist",
        help="The directory to serve builds from.",
    )
    up.add_argument(
        "--bg",
        action="store_true",
        help="Run the index server in the background, exiting after it is ready.",
    )
    up.add_argument(
        "--offline",
        action="store_true",
        help="Run the index server without access to the real PyPI.",
    )
    up.set_defaults(call=_call_index_up)

    down = subparsers.add_parser("down", help="Stop a running package index server.")
    down.set_defaults(call=_call_index_down)

    build = subparsers.add_parser("build", help="Build a static package index server.")
    build.add_argument(
        "targets",
        type=Path,
        nargs="*",
        help="The scenario files to load",
    )
    build.add_argument(
        "--short-names",
        action="store_true",
        help="Exclude scenario names from generated packages.",
    )
    build.add_argument(
        "--no-hash",
        action="store_true",
        help="Exclude scenario version hashes from generated packages.",
    )
    build.add_argument(
        "--dist-dir",
        type=Path,
        default=None,
        help="An existing distribution directory to use. If provided, the scenarios will not be rebuilt.",
    )
    build.set_defaults(call=_call_index_build)

    _add_shared_arguments(parser)
    _add_shared_arguments(up)
    _add_shared_arguments(down)
    _add_shared_arguments(build)


def _add_list_parser(subparsers):
    parser = subparsers.add_parser("list", help="List scenarios at the given paths")
    parser.set_defaults(call=_call_list)
    parser.add_argument(
        "targets",
        type=Path,
        nargs="*",
        help="The scenario files to load",
    )
    parser.add_argument(
        "--no-hash",
        action="store_true",
        help="Do not include in the scenario version hashes in the displayed names.",
    )
    parser.add_argument(
        "--skip-invalid",
        action="store_true",
        help="Skip invalid scenario files instead of failing.",
    )
    parser.add_argument(
        "--no-sources",
        action="store_true",
        help="Do not show the source file for each scenario.",
    )
    parser.add_argument(
        "--short-names",
        action="store_true",
        help="Exclude scenario names from generated packages.",
    )
    _add_shared_arguments(parser)


def _add_inspect_parser(subparsers):
    parser = subparsers.add_parser(
        "inspect", help="Inspect scenarios at the given paths"
    )
    parser.set_defaults(call=_call_inspect)
    parser.add_argument(
        "targets",
        type=Path,
        nargs="*",
        help="The scenario files to load",
    )
    parser.add_argument(
        "--skip-invalid",
        action="store_true",
        help="Skip invalid scenario files instead of failing.",
    )
    parser.add_argument(
        "--short-names",
        action="store_true",
        help="Exclude scenario names from generated packages.",
    )
    parser.add_argument(
        "--no-hash",
        action="store_true",
        help="Do not include in the scenario version hashes in the displayed names.",
    )
    _add_shared_arguments(parser)


def _add_shared_arguments(parser):
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Enable debug logging",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Disable logging",
    )


def _add_fetch_parser(subparsers):
    parser = subparsers.add_parser(
        "fetch", help="Fetch built-in scenarios from the packse repository"
    )
    parser.set_defaults(call=_call_fetch)
    parser.add_argument(
        "--dest",
        type=Path,
        help="Where to place the fetched scenarios",
    )
    parser.add_argument(
        "--ref",
        type=str,
        default=None,
        help="The reference to fetch",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Allow replacement of existing destination directory",
    )
    _add_shared_arguments(parser)


def get_parser() -> argparse.ArgumentParser:
    parser = _root_parser()
    subparsers = parser.add_subparsers(title="commands")
    _add_build_parser(subparsers)
    _add_build_package_parser(subparsers)
    _add_view_parser(subparsers)
    _add_publish_parser(subparsers)
    _add_list_parser(subparsers)
    _add_inspect_parser(subparsers)
    _add_serve_parser(subparsers)
    _add_fetch_parser(subparsers)
    _add_index_parser(subparsers)

    return parser



"""
Tests a subset of included scenarios
"""

import pytest
from packse import __development_base_path__

from .common import snapshot_command

TEST_SCENARIOS = tuple(
    sorted(
        path
        for path in (__development_base_path__ / "scenarios" / "examples").iterdir()
    )
)
TEST_SCENARIO_IDS = tuple(path.with_suffix("").name for path in TEST_SCENARIOS)


@pytest.mark.parametrize(
    "target",
    TEST_SCENARIOS,
    ids=TEST_SCENARIO_IDS,
)
@pytest.mark.usefixtures("tmpcwd")
def test_build_test_scenarios(snapshot, target):
    assert (
        snapshot_command(
            ["build", str(target)], snapshot_stderr=False, snapshot_filesystem=True
        )
        == snapshot
    )


@pytest.mark.parametrize("target", TEST_SCENARIOS, ids=TEST_SCENARIO_IDS)
@pytest.mark.usefixtures("tmpcwd")
def test_view_test_scenarios(snapshot, target):
    assert snapshot_command(["view", str(target)], snapshot_filesystem=True) == snapshot
